@@ -1,11 +1,12 @@
 package Perinci::CmdLine::Inline;
 
-our $DATE = '2015-09-29'; # DATE
-our $VERSION = '0.25'; # VERSION
+our $DATE = '2015-09-30'; # DATE
+our $VERSION = '0.26'; # VERSION
 
 use 5.010001;
 use strict;
 use warnings;
+use Log::Any::IfLOG qw($log);
 
 use Data::Dmp;
 use Module::Path::More qw(module_path);
@@ -24,11 +25,14 @@ sub _deparse {
     Data::Dumper::Dumper($_[0]);
 }
 
-sub _get_module {
-    my $path = module_path(module => $_[0]) or die "Can't load module '$_[0]'";
+sub _add_module {
+    my ($cd, $mod) = @_;
+    return if $cd->{module_srcs}{$mod};
+    $log->infof("Adding source code of module %s ...", $mod);
+    my $path = module_path(module => $mod) or die "Can't load module '$mod'";
     local $/;
     open my($fh), "<", $path or die "Can't read file '$path': $!";
-    ~~<$fh>;
+    $cd->{module_srcs}{$mod} = <$fh>;
 }
 
 # keep synchronize with Perinci::CmdLine::Base
@@ -281,17 +285,18 @@ sub gen_inline_pericmd_script {
         modules => {},
         vars => {},
         subs => {},
-        module_srcs => {
-            map {$_ => _get_module($_)} (
-                "Data::Check::Structure",
-                "Text::Table::Tiny",
-                "Getopt::Long::EvenLess",
-                "Perinci::Result::Format::Lite",
-                @{ $args{include} // [] },
-            )
-        },
+        module_srcs => {},
     };
 
+    for ("Data::Check::Structure",
+         "Getopt::Long::EvenLess",
+         "Perinci::Result::Format::Lite",
+         "Text::Table::Tiny",
+         @{ $args{include} // [] },
+     ) {
+        _add_module($cd, $_);
+    }
+
   GEN_SCRIPT:
     {
         no strict 'refs';
@@ -308,6 +313,7 @@ sub gen_inline_pericmd_script {
             $copts{help} = $Perinci::CmdLine::Base::copts{help};
             $copts{version} = $Perinci::CmdLine::Base::copts{version};
             unless ($skip_format) {
+                _add_module($cd, "JSON::Tiny");
                 $copts{json} = $Perinci::CmdLine::Base::copts{json};
                 $copts{format} = $Perinci::CmdLine::Base::copts{format};
                 # "naked_res!" currently not supported by
@@ -349,13 +355,23 @@ _
     print "\n";
 _
 
-        $cd->{subs}{_pci_json} = <<'_';
-     state $json = do {
-        # XXX try JSON::XS first, fallback to JSON::PP
-        require JSON::PP;
-        JSON::PP->new->canonical(1)->allow_nonref;
+        # not yet needed
+        #        $cd->{subs}{_pci_encode_json} = <<'_';
+        #    state $code = do {
+        #        if    (eval { require JSON::XS; 1 }) { my $json = JSON::XS->new->canonical(1)->allow_nonref; sub { $json->encode(shift) } }
+        #        elsif (eval { require JSON::PP; 1 }) { my $json = JSON::PP->new->canonical(1)->allow_nonref; sub { $json->encode(shift) } }
+        #        else { require JSON::Tiny; \&JSON::Tiny::encode_json }
+        #    };
+        #    $code->(shift);
+        #_
+
+        $cd->{subs}{_pci_decode_json} = <<'_';
+    state $code = do {
+        if    (eval { require JSON::XS; 1 }) { my $json = JSON::XS->new->canonical(1)->allow_nonref; sub { $json->decode(shift) } }
+        elsif (eval { require JSON::PP; 1 }) { my $json = JSON::PP->new->canonical(1)->allow_nonref; sub { $json->decode(shift) } }
+        else { require JSON::Tiny; \&JSON::Tiny::decode_json }
     };
-    $json;
+    $code->(shift);
 _
 
         {
@@ -498,7 +514,8 @@ _
                         if (($specmeta->{parsed}{type} // '') =~ /\@/) {
                             push @l, 'if ($mentioned_args{\'', $specmeta->{arg}, '\'}++) { push @{ $_pci_args{\'', $specmeta->{arg}, '\'} }, $_[1] } else { $_pci_args{\'', $specmeta->{arg}, '\'} = [$_[1]] }';
                         } elsif ($specmeta->{is_json}) {
-                            push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_json->decode($_[1]);';
+                            push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = _pci_decode_json($_[1]);';
+                            _add_module($cd, "JSON::Tiny");
                         } else {
                             push @l, '$_pci_args{\'', $specmeta->{arg}, '\'} = $_[1];';
                         }
@@ -680,7 +697,7 @@ Perinci::CmdLine::Inline - Generate inline Perinci::CmdLine CLI script
 
 =head1 VERSION
 
-This document describes version 0.25 of Perinci::CmdLine::Inline (from Perl distribution Perinci-CmdLine-Inline), released on 2015-09-29.
+This document describes version 0.26 of Perinci::CmdLine::Inline (from Perl distribution Perinci-CmdLine-Inline), released on 2015-09-30.
 
 =head1 SYNOPSIS
 
@@ -704,26 +721,6 @@ This document describes version 0.25 of Perinci::CmdLine::Inline (from Perl dist
 
 B<EARLY DEVELOPMENT.>
 
-Sample startup overhead benchmark vs Perinci::CmdLine::Lite:
-
-                                   Rate run 1 (lite) run --version (lite)    -c (lite) run 1 (inline) run --version (inline) -c (inline)   perl
- run 1 (lite)            12.75+-0.011/s           --               -60.2%       -82.9%         -87.7%                 -91.4%      -93.6% -97.7%
- run --version (lite)   32.067+-0.096/s 151.5+-0.78%                   --       -57.0%         -69.0%                 -78.3%      -83.8% -94.1%
- -c (lite)                74.65+-0.18/s  485.4+-1.5%         132.78+-0.9%           --         -27.9%                 -49.4%      -62.4% -86.3%
- run 1 (inline)          103.58+-0.33/s  712.4+-2.7%            223+-1.4% 38.76+-0.56%             --                 -29.8%      -47.8% -80.9%
- run --version (inline)      147.6+-1/s 1057.9+-8.2%          360.4+-3.5%   97.8+-1.5%     42.5+-1.1%                     --      -25.6% -72.8%
- -c (inline)               198.4+-1.3/s 1456.3+-9.9%          518.8+-4.3%  165.8+-1.8%     91.6+-1.4%             34.4+-1.3%          -- -63.5%
- perl                      543.3+-3.6/s    4161+-28%            1594+-12%  627.8+-5.1%    424.5+-3.8%              268+-3.5% 173.8+-2.5%     --
- 
- Average times:
-   perl                  :     1.8406ms
-   -c (inline)           :     5.0403ms
-   run --version (inline):     6.7751ms
-   run 1 (inline)        :     9.6544ms
-   -c (lite)             :    13.3958ms
-   run --version (lite)  :    31.1847ms
-   run 1 (lite)          :    78.4314ms
-
 =head1 FUNCTIONS
 
 
